#!/usr/bin/env python3

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import transaction
from django.db.models import Q
from django.db.utils import IntegrityError

from arkindex.images.models import ImageServer
from arkindex.ponos.models import Farm

# Constants used in architecture project
UPLOADS_IMAGE_SERVER_ID = 12345
UPLOADS_IMAGE_SERVER_URL = "https://uploads.iiif.ark.localhost/iiif/2"
UPLOADS_IMAGE_SERVER_BUCKET = "uploads"
UPLOADS_IMAGE_SERVER_REGION = "local"
INGEST_IMAGE_SERVER_ID = 67890
INGEST_IMAGE_SERVER_URL = "https://ingest.iiif.ark.localhost/iiif/2"
INGEST_IMAGE_SERVER_BUCKET = "ingest"
INGEST_IMAGE_SERVER_REGION = "local"
PONOS_FARM_ID = "001e411a-1111-2222-3333-444455556666"
PONOS_FARM_NAME = "Bootstrap farm"
PONOS_FARM_SEED = "b12868101dab84984481741663d809d2393784894d6e807ceee0bd95051bf971"


class Command(BaseCommand):
    help = "Build required Database model instances for local development."

    def success(self, msg):
        """Helper to display success messages"""
        self.stdout.write(self.style.SUCCESS(f"✅ {msg}"))

    def warn(self, msg):
        """Helper to display warning messages"""
        self.stdout.write(self.style.WARNING(f"⚠ {msg}"))

    def fail(self, msg):
        """Helper to display error messages"""
        self.stdout.write(self.style.ERROR(f"❌ {msg}"))

    def create_image_server(self, id, url, bucket, region, display_name):
        try:
            server = ImageServer.objects.get(Q(id=id) | Q(url=url))
            if server.id != id:
                # Migrate existing images & server id in a single transaction
                with transaction.atomic():
                    server.images.update(server_id=id)
                    ImageServer.objects.filter(id=server.id).update(id=id)
                self.warn(f"Image server {server.id} updated to {id}")

                # Update internal reference for updates below
                server.id = id

            if server.url != url:
                server.url = url
                server.save()

            # Update base settings
            if server.s3_bucket != bucket or server.s3_region != region:
                server.s3_bucket = bucket
                server.s3_region = region
                server.save()
                self.warn(f"Updated image server {server.id} S3 settings")
            else:
                self.success(f"Image server {server.id} valid")
        except ImageServer.DoesNotExist:
            try:
                server = ImageServer.objects.create(
                    id=id,
                    url=url,
                    s3_bucket=bucket,
                    s3_region=region,
                    display_name=display_name,
                )
                self.success(f"Image server {server.id} created")
            except IntegrityError as e:
                self.fail(f"Failed to create image server: {e}")
                return
        return server

    def handle(self, **options):
        # Never allow running this script in production
        if not settings.DEBUG:
            self.fail("You cannot run this script in production.")
            return

        # a Ponos farm with specific ID & seed
        try:
            farm = Farm.objects.get(id=PONOS_FARM_ID)
            update_farm = False

            if farm.seed != PONOS_FARM_SEED:
                farm.seed = PONOS_FARM_SEED
                update_farm = True

            if farm.name != PONOS_FARM_NAME:
                farm.name = PONOS_FARM_NAME
                update_farm = True

            if update_farm:
                farm.save()
                self.warn(f"Ponos farm {farm.name} updated")
            else:
                self.success("Ponos farm valid")
        except Farm.DoesNotExist:
            Farm.objects.create(
                id=PONOS_FARM_ID,
                name=PONOS_FARM_NAME,
                seed=PONOS_FARM_SEED,
            )
            self.success("Ponos farm created")

        # an image server for local cantaloupe https://ark.localhost/iiif/2
        uploads_server = self.create_image_server(UPLOADS_IMAGE_SERVER_ID , UPLOADS_IMAGE_SERVER_URL, UPLOADS_IMAGE_SERVER_BUCKET , UPLOADS_IMAGE_SERVER_REGION , "Local IIIF server for user uploaded files through frontend")
        if uploads_server is None:
            return
        self.create_image_server(INGEST_IMAGE_SERVER_ID , INGEST_IMAGE_SERVER_URL, INGEST_IMAGE_SERVER_BUCKET , INGEST_IMAGE_SERVER_REGION , "Local IIIF server for ingested files from minio")

        # Check there is not already a local server with invalid path
        # We'll merge its image into the new one
        # This bad server may have been created by automatic IIIF server detection
        try:
            bad_server = ImageServer.objects.get(url="https://uploads.iiif.ark.localhost/iiif")
            bad_server.merge_into(uploads_server)
            self.warn(f"Merged images from {bad_server.id} into {uploads_server.id}")

            bad_server.delete()
            self.warn("Deleted old server")
        except ImageServer.DoesNotExist:
            pass
