from pathlib import Path
from unittest.mock import call, patch, seal

from arkindex.ponos.models import Farm, State
from arkindex.ponos.tasks import run_task_rq
from arkindex.process.models import ProcessMode
from arkindex.project.tests import FixtureTestCase


class TestRunTaskRQ(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.farm = Farm.objects.first()
        cls.process = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            corpus=cls.corpus,
            farm=cls.farm,
        )
        cls.task = cls.process.tasks.create(
            slug="something",
            image="image",
            run=0,
            depth=0,
            state=State.Pending,
            ttl=0,
        )

    def test_no_image(self):
        self.task.image = ""
        self.task.save()

        with self.assertRaisesMessage(ValueError, f"Task {self.task} has no Docker image."):
            run_task_rq(self.task)

    def test_not_pending(self):
        for state in set(State) - {State.Pending}:
            with (
                self.subTest(state=state),
                self.assertRaisesMessage(ValueError, f"Task {self.task} must be in pending state to run in RQ."),
            ):
                self.task.state = state
                self.task.save()
                run_task_rq(self.task)


    def test_non_completed_parents(self):
        self.task.depth = 1
        self.task.save()
        for i, state in enumerate({State.Completed, State.Unscheduled, State.Pending}):
            self.task.parents.create(
                slug=f"parent{i}",
                process=self.process,
                run=0,
                depth=0,
                state=state,
                ttl=0,
            )
        parent = self.task.parents.create(
            slug="parent4",
            process=self.process,
            run=0,
            depth=0,
            ttl=0,
        )
        for state in {State.Stopped, State.Error, State.Failed, State.Cancelled}:
            self.task.state = State.Pending
            self.task.save()
            with self.subTest(state=state):
                parent.state = state
                parent.save()
                run_task_rq(self.task)
                self.task.refresh_from_db()
                self.assertEqual(self.task.state, state)


    @patch("arkindex.ponos.models.TaskLogs.s3_object")
    @patch("arkindex.ponos.tasks.tempfile.TemporaryDirectory")
    @patch("arkindex.ponos.tasks.docker.from_env")
    @patch("arkindex.ponos.tasks.run_docker_task")
    def test_exception_handling(
        self,
        run_docker_task_mock,
        docker_from_env_mock,
        temp_dir_mock,
        logs_s3_object_mock,
    ):
        # Those are passed as arguments of run_docker_task
        docker_from_env_mock.return_value = object()
        seal(docker_from_env_mock)
        temp_dir_mock.return_value.__enter__.return_value = "tempdir"
        temp_dir_mock.return_value.__exit__.return_value = None
        seal(temp_dir_mock)

        run_docker_task_mock.side_effect = ZeroDivisionError
        seal(run_docker_task_mock)

        # After the failure, the exception is appended to the task's logs
        logs_s3_object_mock.load.return_value = None
        logs_s3_object_mock.download_fileobj.side_effect = lambda fileobj: fileobj.write(b"Existing logs")
        logs_s3_object_mock.upload_fileobj.return_value = None
        seal(logs_s3_object_mock)

        with self.assertRaises(ZeroDivisionError):
            run_task_rq(self.task)

        self.task.refresh_from_db()
        self.assertEqual(self.task.state, State.Error)

        self.assertEqual(docker_from_env_mock.call_count, 1)
        self.assertEqual(docker_from_env_mock.call_args, call())
        self.assertEqual(temp_dir_mock.call_count, 1)
        self.assertEqual(temp_dir_mock.call_args, call(suffix=f"_{self.task.id}"))

        self.assertEqual(run_docker_task_mock.call_count, 1)
        self.assertEqual(run_docker_task_mock.call_args, call(docker_from_env_mock.return_value, self.task, Path("tempdir")))

        self.assertEqual(logs_s3_object_mock.load.call_count, 1)
        self.assertEqual(logs_s3_object_mock.load.call_args, call())
        self.assertEqual(logs_s3_object_mock.download_fileobj.call_count, 1)
        self.assertEqual(logs_s3_object_mock.upload_fileobj.call_count, 1)
        (fileobj, ), kwargs = logs_s3_object_mock.upload_fileobj.call_args
        self.assertEqual(fileobj.getvalue(), b"Existing logs\nPonos exception: ZeroDivisionError()")
        self.assertDictEqual(kwargs, {"ExtraArgs": {"ContentType": "text/plain; charset=utf-8"}})
