from unittest.mock import patch

from rest_framework import status
from rest_framework.reverse import reverse

from arkindex.ponos.models import Farm
from arkindex.process.models import ArkindexFeature, Process, ProcessMode, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase


class TestElementsInit(FixtureAPITestCase):
    """
    Test the creation of the elements initialisation task in processes
    """
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        cls.reco_version = WorkerVersion.objects.get(worker__slug="reco")
        cls.dla_version = WorkerVersion.objects.get(worker__slug="dla")
        cls.process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            farm=Farm.objects.first(),
            corpus=cls.corpus
        )

    @patch("arkindex.process.tasks.initialize_activity.delay")
    def test_start_process_with_init_worker_run(self, activities_delay_mock):
        """
        When a process has a worker run using the elements initialisation worker version,
        only one task is created for that worker regardless of chunks.
        """
        self.client.force_login(self.user)

        cases = [(1, {}), (3, {"chunks": 3})]
        for chunks_count, params in cases:
            with self.subTest(chunks_count=chunks_count, params=params):
                process = Process.objects.create(
                    mode=ProcessMode.Workers,
                    creator=self.user,
                    farm=Farm.objects.first(),
                    corpus=self.corpus
                )
                init_run = process.worker_runs.create(
                    version=self.init_elements_version,
                    ttl=0,
                )
                worker_run = process.worker_runs.create(
                    version=self.reco_version,
                    parents=[init_run.id],
                    ttl=0,
                )

                with self.assertNumQueries(16):
                    response = self.client.post(
                        reverse("api:process-start", kwargs={"pk": str(process.id)}),
                        params
                    )
                    self.assertEqual(response.status_code, status.HTTP_201_CREATED)

                process.refresh_from_db()
                self.assertEqual(process.tasks.count(), (1 + chunks_count*1))
                self.assertEqual(process.chunks, chunks_count)

                # One initialisation task
                init_task = process.tasks.get(slug="initialisation")
                self.assertEqual(init_task.command, self.init_elements_version.docker_command)
                self.assertEqual(init_task.image, self.init_elements_version.docker_image_iid)
                self.assertListEqual(list(init_task.parents.all()), [])

                # One or $chunks_count worker task(s)
                worker_tasks = [task for task in process.tasks.order_by("slug") if task.id != init_task.id]
                for i, task in enumerate(worker_tasks):
                    task_slug = worker_run.task_slug
                    if chunks_count > 1:
                        task_slug = f"{task_slug}_{i+1}"
                    self.assertEqual(task.slug, task_slug)
                    self.assertEqual(task.worker_run, worker_run)
                    self.assertListEqual(list(task.parents.all()), [init_task])

    @patch("arkindex.process.tasks.initialize_activity.delay")
    def test_start_process_no_init_worker_run(self, activities_delay_mock):
        """
        If the process does not have any worker runs using the elements initialisation worker version,
        then one is created when the process is started.
        """
        self.client.force_login(self.user)
        reco_run = self.process.worker_runs.create(version=self.reco_version, ttl=0)
        dla_run = self.process.worker_runs.create(version=self.dla_version, parents=[reco_run.id], ttl=0)

        with self.assertNumQueries(18):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.process.refresh_from_db()
        self.assertEqual(self.process.tasks.count(), 3)

        # An elements initialisation worker run has been created for the process, and added as a
        # parent to the parent-less worker runs of that process
        init_run = self.process.worker_runs.get(version_id=self.init_elements_version.id)
        reco_run.refresh_from_db()
        dla_run.refresh_from_db()
        self.assertListEqual(reco_run.parents, [init_run.id])
        self.assertListEqual(dla_run.parents, [reco_run.id])

        init_task = self.process.tasks.get(slug="initialisation")
        self.assertEqual(init_task.command, self.init_elements_version.docker_command)
        self.assertEqual(init_task.image, self.init_elements_version.docker_image_iid)
        self.assertListEqual(list(init_task.parents.all()), [])

        reco_task = self.process.tasks.get(slug=reco_run.task_slug)
        self.assertListEqual(list(reco_task.parents.all()), [init_task])

        dla_task = self.process.tasks.get(slug=dla_run.task_slug)
        self.assertListEqual(list(dla_task.parents.all()), [reco_task])
