from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.process.models import FeatureUsage, Worker
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestWorkerVersionRetrieve(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_reco = Worker.objects.get(slug="reco")
        cls.version_1 = cls.worker_reco.versions.get()

    def test_invalid_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:version-retrieve", kwargs={"pk": "12341234-1234-1234-1234-123412341234"})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(response.json(), {"detail": "No WorkerVersion matches the given query."})

    @patch("arkindex.users.utils.filter_rights", return_value=Worker.objects.none())
    def test_requires_contributor(self, filter_rights_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(response.json(), {"detail": "No WorkerVersion matches the given query."})

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Worker, Role.Contributor.value))

    def test_retrieve(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.version_1.id),
            "configuration": {"test": 42},
            "docker_image_iid": self.version_1.docker_image_iid,
            "state": "available",
            "gpu_usage": "disabled",
            "model_usage": FeatureUsage.Disabled.value,
            "revision_url": None,
            "version": 1,
            "tag": None,
            "branch": None,
            "worker": {
                "id": str(self.worker_reco.id),
                "name": "Recognizer",
                "type": "recognizer",
                "slug": "reco",
                "description": "",
                "archived": False,
                "repository_url": None,
                "cost_1k_elements": "0.250",
                "cost_cpu_hour": "1.312",
                "cost_gpu_hour": "0.000",
            },
            "created": "2020-02-02T01:23:45.678000Z",
        })
