import os
from pathlib import Path
from unittest.mock import patch

from django.conf import settings
from django.core.checks import Critical, Error, Warning
from django.test import TestCase, override_settings
from django.urls import path


@override_settings(ALL_CHECKS=True)
class ChecksTestCase(TestCase):

    def test_api_urls_check(self):
        """
        Check should always be OK
        """
        from arkindex.project.api_v1 import api
        from arkindex.project.checks import api_urls_check

        self.assertListEqual(api_urls_check(), [])

        # Add an invalid url
        bad_url = path("bad", lambda: None, name="bad-url")
        api.append(bad_url)

        # Should trigger the check
        self.assertListEqual(
            api_urls_check(),
            [
                Error(
                    "This API URL is missing a mandatory ending /",
                    hint=str(bad_url),
                    id="arkindex.E001",
                )
            ]
        )

    def test_local_imageserver_check(self):
        """
        Test the local imageserver existence check
        """
        from arkindex.images.models import ImageServer
        from arkindex.project.checks import local_imageserver_check
        self.assertEqual(ImageServer.objects.count(), 0)

        with self.settings(LOCAL_IMAGESERVER_ID=42):
            self.assertListEqual(local_imageserver_check(), [
                Warning(
                    "Local ImageServer with ID 42 does not exist.",
                    hint="settings.LOCAL_IMAGESERVER_ID = 42",
                    id="arkindex.W009",
                ),
            ])

        srv = ImageServer.objects.create(display_name="Test Local Server", url="http://nowhere")

        with self.settings(LOCAL_IMAGESERVER_ID=srv.id):
            self.assertListEqual(local_imageserver_check(), [])

        srv.delete()

    @override_settings()
    def test_ponos_env_check(self):
        from arkindex.project.checks import ponos_env_check
        self.assertListEqual(ponos_env_check(), [])

        settings.CONFIG_PATH = Path("/somewhere/config.yml")
        settings.PONOS_DEFAULT_ENV = {
            "ARKINDEX_API_TOKEN": "oh no",
        }
        self.assertListEqual(ponos_env_check(), [
            Warning(
                "The ARKINDEX_API_URL environment variable should be defined "
                "to allow API client autoconfiguration in Ponos tasks.",
                hint="`ponos.default_env` in /somewhere/config.yml",
                id="arkindex.W006",
            ),
            Warning(
                "The ARKINDEX_API_CSRF_COOKIE environment variable should be defined "
                "to allow API client autoconfiguration in Ponos tasks.",
                hint="`ponos.default_env` in /somewhere/config.yml",
                id="arkindex.W006",
            ),
            Warning(
                "Defining a default value for the ARKINDEX_API_TOKEN variable may introduce a security issue "
                "by allowing Ponos tasks to bypass the Ponos task authentication. Consider removing it.",
                hint="`ponos.default_env.ARKINDEX_API_TOKEN` in /somewhere/config.yml",
                id="arkindex.W013",
            ),
        ])

    @override_settings()
    def test_s3_check(self):
        from arkindex.project.checks import s3_check

        del settings.AWS_ACCESS_KEY
        del settings.AWS_SECRET_KEY
        del settings.AWS_THUMBNAIL_BUCKET
        del settings.AWS_STAGING_BUCKET
        del settings.AWS_EXPORT_BUCKET
        self.assertCountEqual(s3_check(), [
            Error(
                "AWS access key ID is missing; all S3-related features will fail.",
                hint="settings.AWS_ACCESS_KEY = None",
                id="arkindex.E011",
            ),
            Error(
                "AWS secret key is missing; all S3-related features will fail.",
                hint="settings.AWS_SECRET_KEY = None",
                id="arkindex.E011",
            ),
            Error(
                "S3 thumbnails bucket name is missing; all S3-related features will fail.",
                hint="settings.AWS_THUMBNAIL_BUCKET = None",
                id="arkindex.E011",
            ),
            Error(
                "S3 staging bucket name is missing; all S3-related features will fail.",
                hint="settings.AWS_STAGING_BUCKET = None",
                id="arkindex.E011",
            ),
            Error(
                "S3 export bucket name is missing; all S3-related features will fail.",
                hint="settings.AWS_EXPORT_BUCKET = None",
                id="arkindex.E011",
            ),
        ])

        settings.AWS_ACCESS_KEY = "key"
        settings.AWS_SECRET_KEY = "s3kr3t"
        settings.AWS_THUMBNAIL_BUCKET = "Thumbs.db"
        settings.AWS_STAGING_BUCKET = "buckette"
        settings.AWS_EXPORT_BUCKET = "devnull"
        self.assertListEqual(s3_check(), [])

    @override_settings()
    def test_public_hostname_check(self):
        from arkindex.project.checks import public_hostname_check

        settings.PUBLIC_HOSTNAME = None
        self.assertListEqual(public_hostname_check(), [
            Warning(
                "The public_hostname setting is missing; absolute URLs may not be correctly generated.",
                hint="settings.PUBLIC_HOSTNAME",
                id="arkindex.W008",
            )
        ])

        settings.PUBLIC_HOSTNAME = "https://darkindex.lol"
        self.assertListEqual(public_hostname_check(), [])

    @patch("arkindex.project.checks.sys.version_info")
    def test_python_version_check(self, version_info_mock):
        from arkindex.project.checks import python_version_check

        version_info_mock.minor = 9
        self.assertListEqual(python_version_check(), [
            Critical(
                "Arkindex requires Python 3.10 or later.",
                id="arkindex.C001",
            )
        ])

        version_info_mock.minor = 10
        self.assertListEqual(python_version_check(), [])

        version_info_mock.minor = 11
        self.assertListEqual(python_version_check(), [])

    @override_settings()
    def test_ingest_check(self):
        from arkindex.images.models import ImageServer
        from arkindex.project.checks import ingest_check

        settings.ARKINDEX_FEATURES = {"ingest": False}
        settings.INGEST_S3_ACCESS_KEY = None
        settings.INGEST_S3_SECRET_KEY = None
        settings.INGEST_S3_ENDPOINT = None
        settings.INGEST_S3_REGION = None
        settings.INGEST_IMAGESERVER_ID = None
        self.assertListEqual(ingest_check(), [])

        settings.ARKINDEX_FEATURES = {"ingest": True}
        self.assertCountEqual(ingest_check(), [
            Warning(
                "INGEST_S3_ACCESS_KEY is not set. The S3 ingest feature will not work.",
                id="arkindex.W010",
            ),
            Warning(
                "INGEST_S3_SECRET_KEY is not set. The S3 ingest feature will not work.",
                id="arkindex.W010",
            ),
            Warning(
                "Neither INGEST_S3_REGION nor INGEST_S3_ENDPOINT are set. The S3 ingest feature will not work.",
                id="arkindex.W010",
            ),
            Warning(
                "INGEST_IMAGESERVER_ID is not set. The S3 ingest feature will not work.",
                id="arkindex.W010",
            ),
        ])

        settings.INGEST_S3_ACCESS_KEY = "key"
        settings.INGEST_S3_SECRET_KEY = "s3kr3t"
        settings.INGEST_S3_ENDPOINT = "something"
        settings.INGEST_IMAGESERVER_ID = 42
        self.assertListEqual(ingest_check(), [
            Warning(
                "ImageServer with ID 42 does not exist. The S3 ingest feature will not work.",
                hint="settings.INGEST_IMAGESERVER_ID = 42",
                id="arkindex.W010",
            )
        ])

        ImageServer.objects.create(id=42, display_name="Test Server", url="https://null.teklia.com/dev/null")
        self.assertListEqual(ingest_check(), [])

        settings.INGEST_S3_ENDPOINT = None
        settings.INGEST_S3_REGION = "somewhere"
        self.assertListEqual(ingest_check(), [])

    @override_settings()
    @patch("os.path.exists", return_value=True)
    @patch.dict(os.environ, {
        "AWS_ACCESS_KEY_ID": "blah",
        "AWS_CONFIG_FILE": "blah",
        "AWS_CREDENTIAL_FILE": "blah",
        "AWS_DEFAULT_PROFILE": "blah",
        "AWS_DEFAULT_REGION": "blah",
        "AWS_EXECUTION_ENV": "blah",
        "AWS_PROFILE": "blah",
        "AWS_REGION": "blah",
        "AWS_SECRET_ACCESS_KEY": "blah",
        "AWS_SECURITY_TOKEN": "blah",
        "AWS_SESSION_TOKEN": "blah",
        "BOTO_CONFIG": "blah",
    })
    def test_botocore_config_check(self, exists_mock):
        from arkindex.project.checks import botocore_config_check

        expected_warnings = [
            Warning(
                "The AWS_ACCESS_KEY_ID environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The AWS_CONFIG_FILE environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The AWS_CREDENTIAL_FILE environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The AWS_DEFAULT_PROFILE environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The AWS_DEFAULT_REGION environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The AWS_EXECUTION_ENV environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The AWS_PROFILE environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The AWS_REGION environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The AWS_SECRET_ACCESS_KEY environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The AWS_SECURITY_TOKEN environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The AWS_SESSION_TOKEN environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The BOTO_CONFIG environment variable is set, and not all S3 settings are set.\n"
                "This variable might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The ~/.aws/config file exists, and not all S3 settings are set.\n"
                "The settings in this file might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
            Warning(
                "The ~/.aws/credentials file exists, and not all S3 settings are set.\n"
                "The settings in this file might override the Arkindex settings and cause unexpected behavior.",
                id="arkindex.W011",
            ),
        ]

        settings.INGEST_S3_ACCESS_KEY = None
        settings.INGEST_S3_SECRET_KEY = None
        settings.INGEST_S3_ENDPOINT = None
        settings.INGEST_S3_REGION = None
        settings.AWS_ACCESS_KEY = None
        settings.AWS_SECRET_KEY = None
        settings.AWS_ENDPOINT = None
        settings.AWS_REGION = None

        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.INGEST_S3_ACCESS_KEY = "something"
        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.INGEST_S3_SECRET_KEY = "something"
        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.AWS_ACCESS_KEY = "something"
        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.AWS_SECRET_KEY = "something"
        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.AWS_ENDPOINT = "something"
        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.INGEST_S3_ENDPOINT = "something"
        # We have set all the required settings here, so this check is skipped.
        self.assertCountEqual(botocore_config_check(), [])

        # The check requires either the endpoint or the region
        settings.AWS_ENDPOINT = None
        settings.AWS_REGION = "something"
        self.assertCountEqual(botocore_config_check(), [])

        settings.INGEST_S3_ENDPOINT = None
        settings.INGEST_S3_REGION = "something"
        self.assertCountEqual(botocore_config_check(), [])

    def test_signup_default_group(self):
        """
        Test the SIGNUP_DEFAULT_GROUP check
        """
        from arkindex.project.checks import signup_default_group_check
        from arkindex.users.models import Group

        with self.settings(SIGNUP_DEFAULT_GROUP=None):
            self.assertListEqual(signup_default_group_check(), [])

        with self.settings(SIGNUP_DEFAULT_GROUP="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"):
            self.assertListEqual(signup_default_group_check(), [
                Warning(
                    "Group object with ID aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa does not exist",
                    hint="settings.SIGNUP_DEFAULT_GROUP = aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    id="arkindex.W012",
                ),
            ])

        group = Group.objects.create(name="Vocaloid")

        with self.settings(SIGNUP_DEFAULT_GROUP=group.id):
            self.assertListEqual(signup_default_group_check(), [])

    def test_init_elements_worker_version_check_ok(self):
        from arkindex.process.models import ArkindexFeature, Worker, WorkerType, WorkerVersion, WorkerVersionState
        from arkindex.project.checks import init_elements_check

        init_type = WorkerType.objects.create(slug="init", display_name="Elements Initialisation")
        WorkerVersion.objects.create(
            worker=Worker.objects.create(
                name="Elements Initialisation Worker",
                slug="initialisation",
                type=init_type,
            ),
            version=1,
            configuration={
                "docker": {
                    "command": "worker-init-elements"
                }
            },
            state=WorkerVersionState.Available,
            docker_image_iid="registry.teklia.gitlab.com/submarine-reflection:latest",
            feature=ArkindexFeature.InitElements,
        )
        self.assertListEqual(init_elements_check(), [])

    def test_init_elements_worker_version_check_missing(self):
        """
        There is a warning when no worker version for the InitElements feature
        """
        from arkindex.project.checks import init_elements_check

        self.assertListEqual(init_elements_check(), [
            Warning(
                "There is no worker version available to initialize elements.\n"
                "This worker version is required to initialize elements when running a process on Arkindex.\n"
                "Starting or retrying processes will fail.",
                hint="Run `arkindex update_system_workers` or set a worker version's `feature` to `init_elements`.",
                id="arkindex.W014",
            ),
        ])

    @override_settings(ARKINDEX_FEATURES={"ingest": True})
    def test_ingest_worker_version_check_ok(self):
        from arkindex.process.models import ArkindexFeature, Worker, WorkerType, WorkerVersion, WorkerVersionState
        from arkindex.project.checks import ingest_version_check

        import_type = WorkerType.objects.create(slug="import", display_name="Import")
        WorkerVersion.objects.create(
            worker=Worker.objects.create(
                name="S3 Import Worker",
                slug="s3-import",
                type=import_type,
            ),
            version=1,
            state=WorkerVersionState.Available,
            docker_image_iid="registry.teklia.gitlab.com/submarine-reflection:latest",
            feature=ArkindexFeature.S3Ingest,
        )
        self.assertListEqual(ingest_version_check(), [])

    @override_settings(ARKINDEX_FEATURES={"ingest": True})
    def test_ingest_worker_version_check_missing(self):
        """
        There is a warning when no worker version exists for the S3Ingest feature
        """
        from arkindex.project.checks import ingest_version_check

        self.assertListEqual(ingest_version_check(), [
            Warning(
                "There is no worker version available to import elements from S3.\n"
                "Starting or retrying S3 import processes will fail.",
                hint="Run `arkindex update_system_workers` or set a worker version's `feature` to `s3_ingest`.",
                id="arkindex.W015",
            ),
        ])

    @override_settings(ARKINDEX_FEATURES={"ingest": False})
    def test_ingest_worker_version_check_feature_disabled(self):
        """
        The ingest worker version check does not run when the ingest feature flag is disabled
        """
        from arkindex.project.checks import ingest_version_check

        self.assertListEqual(ingest_version_check(), [])

    def test_update_system_workers_check(self):
        """
        The Docker images in the `arkindex update_system_workers` must be manually verified
        as matching the current Arkindex release.
        When the tests run, running the check should never return anything.
        """
        from arkindex.project.checks import update_system_workers_check

        self.assertListEqual(update_system_workers_check(), [])

    @override_settings(VERSION="0.0.0")
    def test_update_system_workers_check_error(self):
        from arkindex.project.checks import update_system_workers_check

        self.assertListEqual(update_system_workers_check(), [Error(
            "The Docker images specified for workers providing features in `update_system_workers` "
            "have not been verified as up to date for Arkindex version 0.0.0.",
            hint="Update `arkindex/system_workers.yml` to the latest versions of the feature workers, "
                 "then update its `version` to '0.0.0'.",
            id="arkindex.E012",
        )])
